"""
A module implementing a protocol factory for the 'fmt ' chunk data.
"""

# third-party
from runtimepy.codec.protocol import Protocol, ProtocolFactory
from runtimepy.enum.registry import EnumRegistry, RuntimeIntEnum
from runtimepy.primitives.byte_order import ByteOrder

ENUMS = EnumRegistry()


class WaveType(RuntimeIntEnum):
    """Enumeration for WAVE data formats."""

    PCM = 1


WaveType.register_enum(ENUMS)


class WaveClass(RuntimeIntEnum):
    """An enumeration for valid kinds of WAVE configurations."""

    MONO_8 = 1
    STEREO_8 = 2
    STEREO_16 = 4


WaveClass.register_enum(ENUMS)
BYTE_ORDER = ByteOrder.LITTLE_ENDIAN


class WaveFormat(ProtocolFactory):
    """Parse the WAVE format data from bytes."""

    protocol: Protocol = Protocol(ENUMS, byte_order=BYTE_ORDER)

    @classmethod
    def initialize(cls, protocol: Protocol) -> None:
        """Initialize this protocol."""

        protocol.add_field("type", "uint16", enum="WaveType")
        protocol.add_field("channels", "uint16")
        protocol.add_field("sample_rate", "uint32")
        protocol.add_field("bytes_per_second", "uint32")

        protocol.add_field("class", "uint16", enum="WaveClass")
        protocol.add_field("bits_per_sample", "uint16")



# =====================================
# generator=datazen
# version=3.1.4
# hash=dd15fcaae00a97a45b4568b47c90f634
# =====================================
"""
quasimoto - Package's default entry-point.
"""

# built-in
import sys

# internal
from quasimoto.entry import main

if __name__ == "__main__":
    sys.exit(main(sys.argv))



"""
A module hosting mixin classes related to WAVE files.
"""

# built-in
from typing import cast

# third-party
from runtimepy.codec.protocol import Protocol
from vcorelib.logging import LoggerMixin

# internal
from quasimoto.wave.protocol import WaveFormat


class FormatMixin(LoggerMixin):
    """A class mixin for classes that use wave format data."""

    byte_order = WaveFormat.protocol.array.byte_order

    def __init__(self) -> None:
        """Initialize this instance."""

        super().__init__()
        self.format = WaveFormat.instance()

    @property
    def channels(self) -> int:
        """Get the number of channels in this stream."""
        return cast(int, self.format["channels"])

    @property
    def sample_bits(self) -> int:
        """Get the number of bits per sample."""
        return cast(int, self.format["bits_per_sample"])

    @property
    def sample_bytes(self) -> int:
        """Get the number of bytes per sample."""

        bits = self.sample_bits
        assert bits % 8 == 0
        return self.sample_bits // 8

    @property
    def sample_rate(self) -> int:
        """Get the sample rate."""
        return cast(int, self.format["sample_rate"])

    @property
    def sample_period(self) -> float:
        """Get the sample period for this data."""
        return 1.0 / self.sample_rate

    def validate_header(self, header: Protocol) -> None:
        """Validate the 'fmt ' chunk data."""

        # Validate bitrate.
        assert (
            header["bytes_per_second"]
            == header["sample_rate"]  # type: ignore
            * header["bits_per_sample"]
            * header["channels"]
            / 8
        )



"""
A module implementing enumeration interfaces for this package.
"""

# built-in
from enum import StrEnum
from typing import BinaryIO, Optional


class AudioFileTypes(StrEnum):
    """An enumeration for supported file types."""

    WAVE = "wav"


DEFAULT_FORMAT = AudioFileTypes.WAVE


class ChunkType(StrEnum):
    """An enumeration for different kinds of RIFF chunks."""

    RIFF = "RIFF"
    LIST = "LIST"
    WAVE = "WAVE"

    FMT = "fmt "
    DATA = "data"
    ID3 = "ID3 "

    @property
    def is_container(self) -> bool:
        """Whether or not this is a container chunk type."""

        return self is ChunkType.RIFF or self is ChunkType.LIST

    @staticmethod
    def from_stream(stream: BinaryIO) -> Optional["ChunkType"]:
        """Read the chunk type from a stream."""

        result = None

        check = stream.read(3).decode("ascii")
        if len(check) == 3:
            # Some files hackily have some 'ID3' metadata at the end?
            if check != "ID3":
                result = ChunkType(check + stream.read(1).decode("ascii"))

        return result

    def to_stream(self, stream: BinaryIO) -> None:
        """Write the chunk header."""

        data = bytes(str(self).encode("ascii"))
        assert len(data) == 4
        stream.write(data)



"""
A module implementing sampler interfaces.
"""

# built-in
from collections.abc import Iterable, Iterator
from copy import copy
import math
from typing import TypeVar

# third-party
from runtimepy.primitives import Double

# internal
from quasimoto.wave.writer import DEFAULT_BITS, DEFAULT_SAMPLE_RATE

DEFAULT_FREQUENCY = 261.63
T = TypeVar("T", bound="Sampler")


class Sampler(Iterable[int]):
    """A base class for iterable sampler interfaces."""

    def __init__(
        self,
        num_bits: int = DEFAULT_BITS,
        sample_rate: int = DEFAULT_SAMPLE_RATE,
        duration_s: float = None,
        frequency: float = DEFAULT_FREQUENCY,
        time: float = 0.0,
        amplitude: float = 1.0,
    ) -> None:
        """Initialize this instance."""

        # Can be changed after initialization.
        self.frequency = Double(value=frequency)
        self.amplitude = Double(value=amplitude)
        self.duration_s = duration_s

        # Runtime state.
        self.time = time

        # Constants / final.
        self.sample_rate = sample_rate
        self.period = 1.0 / self.sample_rate
        # Note: this assumed signed + zero-centered.
        self.num_bits = num_bits
        self.scalar = (2 ** (self.num_bits - 1)) - 1

    def __copy__(self: T) -> T:
        """Create a copy of this instance."""

        return type(self)(
            num_bits=self.num_bits,
            sample_rate=self.sample_rate,
            duration_s=self.duration_s,
            frequency=self.frequency.value,
            time=self.time,
        )

    def harmonic(self, index: int) -> float:
        """Get a harmonic frequency based on this instance's frequency."""
        return float(2**index) * self.frequency.value

    def copy(self: T, harmonic: int = None, duration_s: float = None) -> T:
        """Get a copy of this instance."""

        result = copy(self)

        if duration_s is not None:
            result.duration_s = duration_s
        if harmonic is not None:
            result.frequency.value = result.harmonic(harmonic)

        return result

    def advance(self, steps: int = 1) -> bool:
        """Advance time forward."""

        result = True

        for _ in range(steps):
            self.time += self.period
            if self.duration_s is not None:
                result = self.time < self.duration_s

        return result

    def sin(self, now: float) -> int:
        """Get a raw sin value sample."""

        return int(
            self.scalar
            * self.amplitude.value
            * math.sin(math.tau * now * self.frequency.value)
        )

    def value(self, now: float) -> int:
        """Get the next value."""
        return self.sin(now)

    def __iter__(self) -> Iterator[int]:
        """Return an iterator."""
        return self

    def __next__(self) -> int:
        """Get the next value from this sampler."""

        val: int = self.value(self.time)

        if not self.advance():
            raise StopIteration

        return val



"""
An entry-point for the 'gen' command.
"""

# built-in
import argparse
from pathlib import Path

# third-party
from vcorelib.args import CommandFunction

# internal
from quasimoto import PKG_NAME
from quasimoto.enums import DEFAULT_FORMAT
from quasimoto.riff import RiffInterface


def gen_cmd(args: argparse.Namespace) -> int:
    """Execute the arbiter command."""

    with RiffInterface.from_path(args.output) as writer:
        print(writer)

    return 0


def add_gen_cmd(parser: argparse.ArgumentParser) -> CommandFunction:
    """Add gen-command arguments to its parser."""

    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        default=f"{PKG_NAME}.{DEFAULT_FORMAT}",
        help="output file to write",
    )

    return gen_cmd



"""
A module implementing a RIFF-chunk interface.
"""

# built-in
from typing import NamedTuple, Optional

# internal
from quasimoto.enums import ChunkType


class Chunk(NamedTuple):
    """A container for chunk data."""

    kind: ChunkType
    size: int
    data: Optional[bytes] = None
    form: Optional[ChunkType] = None

    def __str__(self) -> str:
        """Get this chunk as a string."""
        result = f"'{self.kind}' size={self.size}"

        if self.form is not None:
            result += f" (form='{self.form}')"

        return result


NULL_BYTE = "\0".encode()



"""
A module implementing interfaces for reading WAVE files.
"""

# built-in
from contextlib import contextmanager
from io import BytesIO
from pathlib import Path
from typing import Iterator, cast

# third-party
from runtimepy.primitives import Int16
from vcorelib.math.time import nano_str

# internal
from quasimoto.enums import ChunkType
from quasimoto.riff import RiffInterface
from quasimoto.riff.chunk import Chunk
from quasimoto.wave.mixins import FormatMixin


class WaveReader(FormatMixin):
    """A class for reading and writing WAVE files."""

    def __init__(self, riff: RiffInterface) -> None:
        """Initialize this instance."""

        super().__init__()

        assert not riff.is_writer

        # Only expecting 'fmt ' and 'data' chunks.
        chunks = list(riff.chunks())
        assert len(chunks) == 2

        # Parse format.
        format_chunk: Chunk = chunks[0]
        assert format_chunk.kind is ChunkType.FMT
        assert format_chunk.size == 16
        assert format_chunk.data is not None
        self.format.array.update(format_chunk.data)

        # Validate format.
        self.validate_header(self.format)
        self.logger.info("Format header: %s.", self.format)

        # Validate data chunk.
        self.data: Chunk = chunks[1]
        assert self.data.kind is ChunkType.DATA

        # Dump some information.
        self.logger.info("%s of sample data.", self.duration_str)

    @property
    def num_samples(self) -> int:
        """Get the number of samples contained."""

        all_channels = self.channels * self.sample_bytes
        assert self.data.size % all_channels == 0
        return self.data.size // all_channels

    @property
    def duration_s(self) -> float:
        """Get the duration in seconds of this data."""
        return self.num_samples / self.sample_rate

    @property
    def duration_str(self) -> str:
        """Get this data's duration as a human-readable string."""
        return nano_str(int(self.duration_s * 1e9), is_time=True) + "s"

    @property
    def samples(self) -> Iterator[tuple[int, ...]]:
        """Get raw samples as a generator."""

        assert self.data.data is not None

        # Only support reading 16-bit samples.
        assert self.sample_bytes == 2

        num_channels = self.channels

        with BytesIO(self.data.data) as stream:
            with self.log_time("Processing samples", reminder=True):
                for _ in range(self.num_samples):
                    yield tuple(
                        cast(
                            int,
                            Int16.kind.read(
                                stream, byte_order=self.byte_order
                            ),
                        )
                        for _ in range(num_channels)
                    )

                # Confirm we're at the end of the data.
                assert stream.tell() == self.data.size

    @staticmethod
    @contextmanager
    def from_path(path: Path) -> Iterator["WaveReader"]:
        """Get a WAVE reader from a path."""
        with RiffInterface.from_path(path, is_writer=False) as riff:
            yield WaveReader(riff)



# =====================================
# generator=datazen
# version=3.1.4
# hash=b8d2bc2e3ccbd1d7eab451040c2edb8e
# =====================================

"""
A module aggregating package commands.
"""

# built-in
from typing import List as _List
from typing import Tuple as _Tuple

# third-party
from vcorelib.args import CommandRegister as _CommandRegister

# internal
from quasimoto.commands.gen import add_gen_cmd


def commands() -> _List[_Tuple[str, str, _CommandRegister]]:
    """Get this package's commands."""

    return [
        (
            "gen",
            "generate audio",
            add_gen_cmd,
        ),
        ("noop", "command stub (does nothing)", lambda _: lambda _: 0),
    ]



"""
A module implementing interfaces for writing WAVE files.
"""

# built-in
from contextlib import contextmanager
import os
from pathlib import Path
from typing import BinaryIO, Iterable, Iterator

# third-party
from runtimepy.primitives import Int16

# internal
from quasimoto.enums import ChunkType
from quasimoto.riff import RiffInterface
from quasimoto.riff.chunk import Chunk
from quasimoto.wave.mixins import FormatMixin

DEFAULT_SAMPLE_RATE = 44100
DEFAULT_CHANNELS = 2
DEFAULT_BITS = 16


class WaveWriter(FormatMixin):
    """A class for reading and writing WAVE files."""

    def __init__(
        self,
        riff: RiffInterface,
        num_channels: int = DEFAULT_CHANNELS,
        sample_rate: int = DEFAULT_SAMPLE_RATE,
        bits_per_sample: int = DEFAULT_BITS,
    ) -> None:
        """Initialize this instance."""

        super().__init__()
        self.riff = riff
        assert self.riff.is_writer

        # Finish writing RIFF header.
        ChunkType.WAVE.to_stream(self.riff.stream)

        assert (num_channels * bits_per_sample) % 8 == 0
        class_num = num_channels * bits_per_sample // 8

        # Write 'fmt ' chunk.
        self.format["type"] = "pcm"
        self.format["channels"] = num_channels
        self.format["sample_rate"] = sample_rate
        self.format["bytes_per_second"] = int(class_num * sample_rate)
        self.format["class"] = class_num
        self.format["bits_per_sample"] = bits_per_sample

        data = bytes(self.format.array)
        self.riff.write(Chunk(ChunkType.FMT, len(data), data=data))

        # Write 'data' chunk header.
        ChunkType.DATA.to_stream(self.riff.stream)

    @classmethod
    def to_bytes(cls, value: int) -> bytes:
        """Convert a sample value to bytes."""

        # Note the underlying type assumption.
        return Int16.kind.encode(value, byte_order=cls.byte_order)

    @classmethod
    def to_stream(cls, stream: BinaryIO, value: int) -> int:
        """Write a value to a stream."""

        data = cls.to_bytes(value)
        stream.write(data)
        return len(data)

    def write(self, samples: Iterable[tuple[int, ...]]) -> None:
        """Write samples to the output."""

        # Only support writing 16-bit samples.
        assert self.sample_bytes == 2

        with self.log_time("Writing samples", reminder=True):
            self.riff.stream.seek(0, os.SEEK_END)
            size_pos = self.riff.stream.tell()
            self.riff.write_size(0)

            size = 0
            for sample in samples:
                for point in sample:
                    size += self.to_stream(self.riff.stream, point)

            self.riff.write_size(size, seek=size_pos)

    @staticmethod
    @contextmanager
    def from_path(path: Path, **kwargs) -> Iterator["WaveWriter"]:
        """Get a WAVE reader from a path."""
        with RiffInterface.from_path(path) as riff:
            yield WaveWriter(riff, **kwargs)



"""
A module implementing interfaces for RIFF files.
"""

# built-in
from contextlib import contextmanager
import os
from pathlib import Path
from typing import BinaryIO, Iterator, Optional, Type, TypeVar, cast

# third-party
from runtimepy.primitives import Uint32
from runtimepy.primitives.byte_order import ByteOrder
from vcorelib.logging import LoggerMixin

# internal
from quasimoto.enums import ChunkType
from quasimoto.riff.chunk import NULL_BYTE, Chunk

T = TypeVar("T", bound="RiffInterface")


class RiffInterface(LoggerMixin):
    """A class for reading and writing RIFF files."""

    def __init__(self, stream: BinaryIO, is_writer: bool = True) -> None:
        """Initialize this instance."""

        super().__init__()

        self.stream = stream

        # Write the header.
        self.is_writer = is_writer
        if self.is_writer:
            ChunkType.RIFF.to_stream(self.stream)
            # Leave a placeholder for actual size.
            self.write_size(0)
        else:
            header = self.read()
            assert header is not None
            self.header: Chunk = header
            self.logger.info("Header: %s.", self.header)
            assert self.header.kind is ChunkType.RIFF

    def read_size(self) -> int:
        """Read a size from the stream."""

        return cast(
            int,
            Uint32.kind.read(self.stream, byte_order=ByteOrder.LITTLE_ENDIAN),
        )

    def read(self) -> Optional[Chunk]:
        """Read the next chunk."""

        result = None

        print(self.stream.tell())
        kind = ChunkType.from_stream(self.stream)
        if kind is not None:
            size = self.read_size()
            data = None
            form = None

            if not kind.is_container:
                data = self.stream.read(size)
                if size % 2 == 1:
                    self.stream.read(1)  # pragma: nocover
            else:
                form = ChunkType.from_stream(self.stream)

            result = Chunk(kind, size, data=data, form=form)

        return result

    def chunks(self) -> Iterator[Chunk]:
        """Read file chunks."""

        result = self.read()
        while result is not None:
            yield result
            result = self.read()

    def write_size(self, size: int, seek: int = None) -> None:
        """An interface for writing a size field."""

        # Validate size.
        prim = Uint32.kind
        bounds = prim.int_bounds
        assert bounds is not None
        assert bounds.validate(size), size

        if seek is not None:
            self.stream.seek(seek)

        # Write size.
        Uint32.kind.write(
            size, self.stream, byte_order=ByteOrder.LITTLE_ENDIAN
        )

    def _write_data(self, data: bytes) -> None:
        """Write chunk data."""

        size = len(data)
        self.write_size(size)

        # Write data.
        self.stream.write(data)
        if size % 2 == 1:
            self.stream.write(NULL_BYTE)  # pragma: nocover

    def write(self, chunk: Chunk) -> None:
        """Write a chunk to the file."""

        assert self.is_writer

        # Can't write container chunks this way.
        assert not chunk.kind.is_container

        # Write header.
        chunk.kind.to_stream(self.stream)

        # Write data.
        if chunk.data is not None:
            self._write_data(chunk.data)

    def finalize(self) -> None:
        """Finalize the header size."""

        if self.is_writer:
            self.stream.seek(0, os.SEEK_END)
            size = self.stream.tell() - 8
            self.write_size(size, seek=4)
        else:
            remaining = self.stream.read()
            if remaining:
                self.logger.warning(
                    "%d bytes remaining in file!", len(remaining)
                )

    @classmethod
    @contextmanager
    def from_path(
        cls: Type[T], path: Path, is_writer: bool = True
    ) -> Iterator[T]:
        """Create a RIFF interface from a path."""

        with path.open("wb" if is_writer else "rb") as out_fd:
            result = cls(out_fd, is_writer=is_writer)
            yield result
            result.finalize()
